"""

A state based environment that focuses on progressing across states.

"""

from .state_environment import StateBasedEnvironment


METTA_ENVIRONMENT_STATEPROGRESSING_PLUGIN_ID = "metta_stateprogressing_environment"
""" Metta plugin id for the state-progressing environment plugin."""


class StateProgressingEnvironment(StateBasedEnvironment):
    """A state environment implementation that progressed through states.

    This extension of the state based environment, focuses on the state fixtures
    as a linear set, which is to be progressed through.

    The set-state is never allowed to move backwards in the list, and if asked
    progress past states, it will activate them before moving on.

    """



"""

Common METTA plugins and functionality.

Package for common shared Metta plugins that can be used by various
other plugins as a based.

"""
from typing import Dict, Any

from configerus import Config
from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment, METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from .state_environment import (
    StateBasedEnvironment,
    METTA_ENVIRONMENT_STATE_PLUGIN_ID,
)
from .state import (
    EnvironmentStatePlugin,
    METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE,
    METTA_STATE_DEFAULT_PLUGIN_ID,
)
from .labelactivate_state import (
    EnvironmentLabelActivateStatePlugin,
    METTA_STATE_LABELACTIVATE_PLUGIN_ID,
)


@Factory(
    plugin_id=METTA_ENVIRONMENT_STATE_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT],
)
def state_environment_factory(
    config: Config, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
) -> StateBasedEnvironment:
    """Build a FixtureBuilderEnvironment environment."""
    return StateBasedEnvironment(config=config, instance_id=instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_STATE_DEFAULT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE],
)
def default_environment_state_factory(
    environment: Environment, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
) -> EnvironmentStatePlugin:
    """Build a EnvironmentState environment."""
    return EnvironmentStatePlugin(
        environment=environment, instance_id=instance_id, label=label, base=base
    )


@Factory(
    plugin_id=METTA_STATE_LABELACTIVATE_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE],
)
def labelactivate_environment_state_factory(
    environment: Environment, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
) -> EnvironmentLabelActivateStatePlugin:
    """Build a EnvironmentState environment."""
    return EnvironmentLabelActivateStatePlugin(
        environment=environment, instance_id=instance_id, label=label, base=base
    )


# ----- METTA bootstraps that we will use on config objects -----


# pylint: disable=unused-argument
def bootstrap_bootstrapper(config: Config):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    config (Config) : an confiug which can be modified to bootstrap

    """


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

A state that uses labels to manage fixtures on activation.

"""
from typing import List, Dict, Any, Union
from logging import getLogger

from configerus.loaded import Loaded

from mirantis.testing.metta.fixture import Fixture
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD
from mirantis.testing.metta_health.healthcheck import METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK

from .state import EnvironmentStatePlugin


logger = getLogger("metta_state:activatingstate")

METTA_STATE_LABELACTIVATE_PLUGIN_ID = "metta_state_labelactivate"
""" Metta plugin id for the label activate environment state plugin."""

METTA_STATE_FIXTURE_LABEL_ACTIVATE = "state-activate"
"""Metta fixture label that indicates that this fixture should activate on a state."""

METTA_STATE_FIXTURE_LABEL_DEACTIVATE = "state-deactivate"
"""Metta fixture label that indicates that this fixture should deactivate on a state."""


class EnvironmentLabelActivateStatePlugin(EnvironmentStatePlugin):
    """A State piece of a StateBasedEnvironment that processes fixture labels on activate."""

    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return Dict of introspective information about the plugin."""
        state_config: Loaded = self.config().load(self._config_label)

        plugin_info: Dict[str, Any] = super().info(deep=deep)
        plugin_info["activate"] = state_config.get(
            [self._config_base, METTA_STATE_FIXTURE_LABEL_ACTIVATE], default="NONE"
        )

        return plugin_info

    def activate(self):
        """Activate the state plugin.

        This means using config data for the state plugin, and labels for the
        contained fixtures for taking actions on them.
        """
        # parent activate to load fixtures/config
        super().activate()

        # Method 1: activate any fixtures with labels directing activation on
        #    this state.
        self._activate_fixtures_with_labels()

        # method 2: activate any fixtures based on configuration from the state.
        self._activate_fixture_using_state_config()

    def _activate_fixture_using_state_config(self):
        """Use state config to decide to activate fixtures."""
        errors: List[Exception] = []

        state_config: Loaded = self.config().load(self._config_label)

        fixtures_to_activate: List[Union[str, Dict[str, str]]] = state_config.get(
            [self._config_base, METTA_STATE_FIXTURE_LABEL_ACTIVATE], default=[]
        )
        try:
            iter(fixtures_to_activate)
        except TypeError as err:
            raise RuntimeError(
                f"Fixtures to activate was not an iterator of instance_ids: {fixtures_to_activate}"
            ) from err

        for fixture_def in fixtures_to_activate:
            if isinstance(fixture_def, str):
                fixture_def = {"instance_id": fixture_def}
            fixture = self.fixtures().get(**fixture_def)

            try:
                _activate_fixture(fixture)

            # pylint: disable=broad-except
            except Exception as err:
                logger.warning(
                    "Failed checking health %s:%s : %s",
                    fixture.plugin_id,
                    fixture.instance_id,
                    err,
                )
                errors.append(
                    RuntimeError(
                        f"Failed checking fixture health {fixture.plugin_id}:{fixture.instance_id}"
                        f" : {err}"
                    )
                )

        if len(errors) > 0:
            raise RuntimeError(
                "Environment state encountered errors while activating on state labels.", errors
            )

    def _activate_fixtures_with_labels(self):
        """Activate contained fixtures if they have the right labels."""
        errors: List[Exception] = []
        for fixture in self.fixtures().filter(
            labels={METTA_STATE_FIXTURE_LABEL_ACTIVATE: self.instance_id()},
            exception_if_missing=False,
        ):
            try:
                _activate_fixture(fixture)

            # pylint: disable=broad-except
            except Exception as err:
                logger.warning(
                    "Failed checking health %s:%s : %s",
                    fixture.plugin_id,
                    fixture.instance_id,
                    err,
                )
                errors.append(
                    RuntimeError(
                        f"Failed checking fixture health {fixture.plugin_id}:{fixture.instance_id}"
                        f" : {err}"
                    )
                )

        if len(errors) > 0:
            raise RuntimeError(
                "Environment state encountered errors while activating on labels.", errors
            )


def _activate_fixture(fixture: Fixture):
    """Activate a passed fixture, depending on interfaces.

    Returns:
    --------
    Exception that was thrown or None
    """
    # Provisioner and Workload plugins are activated by running "apply"
    if (
        METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER in fixture.interfaces
        or METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD in fixture.interfaces
    ):
        logger.warning("Activating fixture %s:%s", fixture.plugin_id, fixture.instance_id)
        fixture.plugin.prepare()
        fixture.plugin.apply()

    elif METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK in fixture.interfaces:
        logger.warning("Checking fixture health %s:%s", fixture.plugin_id, fixture.instance_id)
        fixture.plugin.health()

    else:
        raise ValueError(
            f"Fixture '{fixture.instance_id} was not activated because it "
            "had no activatable interface."
        )



"""

An environment that operates from a set of states.

The idea here is to provide functionality for progression of an environment
across a set of states.  For this, a state is an environment definition itself;
a container of config and fixtures, themselves contained in a environment.

The code here is not very elegant, but what we do is we provide an environment
plugin which will create a number of state plugins which can be activated. If
an environment is asked for config of fixtures, it retrieves them fromt the
currently active state.  You can ask the environment to switch to a different
state.

Some foolishness is in place for the initial construction, and the current
implementation doesn't share anythin across states except initial config copy()
but the concept works.

As Environments are effectively declarative, this initial functionality does not
restrict state changes (like forcing a forward only progression) but that could
be added if we decide that this code is worth keeping.


I think that these state plugins are not safe to generate across threads at the
same time. This comes down to the need to copy/duplicate the config option.

"""
from typing import Dict, Any
from logging import getLogger

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.fixture import (
    Fixtures,
    METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
)
from mirantis.testing.metta.environment import (
    Environment,
    FixtureBuilderEnvironment,
    METTA_ENTRYPOINT_BOOTSTRAP_ENVIRONMENT,
)
from mirantis.testing.metta.building import (
    FixtureBuildingFromConfigMixin,
    FixtureBuildingFromDictMixin,
)
from mirantis.testing.metta.config import (
    add_config_sources_from_config,
    METTA_CONFIG_CONFIG_SOURCE_KEY,
)
from mirantis.testing.metta.importing import add_imports_from_config, METTA_IMPORT_CONFIG_LABEL
from mirantis.testing.metta.setuptools import (
    setuptools_entrypoint,
    METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY,
)

logger = getLogger("metta_state.defaultstate")

METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE = "state"
"""Plugin interface to behave as an environment state in a StateBasedEnvironment."""

METTA_ENTRYPOINT_BOOTSTRAP_STATE = "metta.bootstraps.states"
"""Setup tools entrypoints for bootstrapping a state object."""

METTA_FIXTURES_CONFIG_STATES_KEY = "states"
"""Configerus .get() base key for retrieving states from config."""


METTA_STATE_DEFAULT_PLUGIN_ID = "metta_state_default"
""" Metta plugin id for the standard environment state plugin."""


# we use the parent methods, but override the constructor.
# pylint: disable=super-init-not-called
class EnvironmentStatePlugin(FixtureBuilderEnvironment):
    """A State piece of a StateBasedEnvironment."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = METTA_FIXTURES_CONFIG_STATES_KEY,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Run the super constructor but also set class properties.

        Interpret provided config and configure the object with all of the
        needed pieces for executing terraform commands.

        This repeats a lof of the Environment functionality as an override,
        mainly so that we can change some small details about plugin generation.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config: Config = environment._config
        """Config object, overridden in activate()."""

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base
        """ configerus get key that should contain all tf config """

        self._fixtures: Fixtures = Fixtures()
        """Children fixtures, typically just the client plugin."""

    # Use State bootstrappers as well as env bootastrappers
    def activate(self):
        """Respond to the state being activated."""
        logger.debug("Default state plugin activated: %s", self.instance_id())

        # Use the protected config so that we don't use a state config by accident
        # pylint: disable=protected-access`
        self._config = self._config.copy()
        Environment.__init__(self, config=self._config, instance_id=self._instance_id)
        FixtureBuildingFromConfigMixin.__init__(
            self, config=self._config, builder_callback=self.new_fixture
        )
        FixtureBuildingFromDictMixin.__init__(self, builder_callback=self.new_fixture)

        # If we received any config directiosn, then we self-bootstrap from the config
        if self._config_label:
            # 1. add any config sources specified in config
            # (Here config tells us to load more config. It is weird but ok.)
            add_config_sources_from_config(
                config=self._config,
                label=self._config_label,
                base=[self._config_base, METTA_CONFIG_CONFIG_SOURCE_KEY],
            )

            # 2. import any python code requested, to make sure that all funxtionality
            #    is in scope that is needed.
            #    Here we say look in the "metta" config for a "imports" section.
            add_imports_from_config(
                config=self._config,
                label=self._config_label,
                base=[self._config_base, METTA_IMPORT_CONFIG_LABEL],
            )

            # 3. run an setuptools bootstrap entrypoints
            env_config = self._config.load(self._config_label)
            environment_boostraps = env_config.get(
                [self._config_base, METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY], default=[]
            )
            setuptools_entrypoint(
                entrypoint=METTA_ENTRYPOINT_BOOTSTRAP_ENVIRONMENT,
                entries=environment_boostraps,
                args=[self],
                kwargs={},
            )

            # 4. build fixtures from config
            labels: Dict[str, str] = {
                "container": "state",
                "container_id": self.instance_id(),
                "state": self.instance_id(),
                "environment": self._environment.instance_id(),
            }
            self.add_fixtures_from_config(
                label=self._config_label,
                base=[self._config_base, METTA_FIXTURES_CONFIG_FIXTURES_LABEL],
                labels=labels,
            )

    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict plugin info."""
        state_info = {
            "name": self.instance_id(),
            # "boostraps": self._environment_boostraps,
        }

        if deep:
            state_info["fixtures"] = self._fixtures.info(deep=deep)

        return state_info



"""

An environment that operates from a set of states.

The idea here is to provide functionality for progression of an environment
across a set of states.  For this, a state is an environment definition itself;
a container of config and fixtures, themselves contained in a environment.

The code here is not very elegant, but what we do is we provide an environment
plugin which will create a number of state plugins which can be activated. If
an environment is asked for config of fixtures, it retrieves them fromt the
currently active state.  You can ask the environment to switch to a different
state.

Some foolishness is in place for the initial construction, and the current
implementation doesn't share anythin across states except initial config copy()
but the concept works.

As Environments are effectively declarative, this initial functionality does not
restrict state changes (like forcing a forward only progression) but that could
be added if we decide that this code is worth keeping.

"""
from typing import Dict, Any

from configerus.config import Config
from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.fixture import (
    Fixture,
    Fixtures,
)
from mirantis.testing.metta.environment import (
    FixtureBuilderEnvironment,
)


METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE = "state"
"""Plugin interface to behave as an environment state in a StateBasedEnvironment."""

METTA_ENTRYPOINT_BOOTSTRAP_STATE = "metta.bootstraps.states"
"""Setup tools entrypoints for bootstrapping a state object."""

METTA_FIXTURES_CONFIG_STATES_KEY = "states"
"""Configerus .get() base key for retrieving states from config."""


METTA_ENVIRONMENT_STATE_PLUGIN_ID = "metta_state_environment"
""" Metta plugin id for the state based environment plugin."""


class StateBasedEnvironment(FixtureBuilderEnvironment):
    """An environment implementation that keeps states."""

    def __init__(
        self, config: Config, instance_id: str, label: str = "", base: str = LOADED_KEY_ROOT
    ):
        """Create the StateBased environment components."""
        self._active_state_id: str = ""
        """Currently active state fixture/plugin."""
        FixtureBuilderEnvironment.__init__(
            self, config=config, instance_id=instance_id, label=label, base=base
        )

        # Build all of the state plugins
        labels: Dict[str, str] = {
            "container": "environment",
            "container_id": self.instance_id(),
            "environment": self.instance_id(),
        }

        try:
            self.add_fixtures_from_config(
                label=label,
                base=[base, METTA_FIXTURES_CONFIG_STATES_KEY],
                labels=labels,
                exception_if_missing=True,
            )

        except KeyError as err:
            raise RuntimeError(f"State environment could't create states: {err}") from err

        # Activate the highest priority state plugin
        self.set_state(
            self._fixtures.get(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE]
            ).instance_id
        )

    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Return dict plugin info."""
        return {
            "name": self.instance_id(),
            "boostraps": self._environment_boostraps,
            "active_state": self._active_state_id,
            "states": self._fixtures.filter(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE]
            ).info(deep=deep),
        }

    def set_state(self, state: str):
        """Change to a different active state."""
        try:
            state_fixture: Fixture = self._fixtures.get(
                instance_id=state, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE]
            )

        except KeyError as err:
            raise RuntimeError(f"Unknown State requested for activation: {state}") from err

        self._active_state_id = state_fixture.instance_id

        # activate the state plugin
        if hasattr(state_fixture.plugin, "activate"):
            state_fixture.plugin.activate()

    def _get_active_state_fixture(self) -> Fixture:
        """Get the active state fixture."""
        return self._fixtures.get(
            instance_id=self._active_state_id,
            interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENTSTATE],
        )

    def config(self) -> Config:
        """Return the Config from the active state."""
        if self._active_state_id:
            return self._get_active_state_fixture().plugin.config()
        return self._config

    def fixtures(self) -> Fixtures:
        """Return the Fixtures from the active state."""
        if self._active_state_id:
            return self._get_active_state_fixture().plugin.fixtures()
        return self._fixtures
